/***************************************************************************
 * Copyright (c)  2014  Zephyr Software LLC. All rights reserved.
 *
 * This software is furnished under a license and/or other restrictive
 * terms and may be used and copied only in accordance with such terms
 * and the inclusion of the above copyright notice. This software or
 * any other copies thereof may not be provided or otherwise made
 * available to any other person without the express written consent
 * of an authorized representative of Zephyr Software LCC. Title to,
 * ownership of, and all rights in the software is retained by
 * Zephyr Software LCC.
 *
 * Zephyr Software LLC. Proprietary Information
 *
 * Unless otherwise specified, the information contained in this
 * directory, following this legend, and/or referenced herein is
 * Zephyr Software LLC. (Zephyr) Proprietary Information. 
 *
 * CONTACT
 *
 * For technical assistance, contact Zephyr Software LCC. at:
 *      
 *
 * Zephyr Software, LLC
 * 2040 Tremont Rd
 * Charlottesville, VA 22911
 *
 * E-mail: jwd@zephyr-software.com
 **************************************************************************/

#ifndef zipr_sdk_zipr_plugin_options_h
#define zipr_sdk_zipr_plugin_options_h

#include <string>
#include <unistd.h>
#include <libIRDB-core.hpp>
#include <iostream>
#include <sstream>

namespace Zipr_SDK
{

enum ZiprOptionType {
	ZiprUntypedOptionType,
	ZiprIntegerOptionType,
	ZiprDoubleOptionType,
	ZiprBooleanOptionType,
	ZiprStringOptionType,
};

class ZiprOption_t 
{
	public:
	/*
		ZiprOption_t(ZiprOptionType type)
			: m_key(""),
			  m_untyped_value(""),
				m_observee(NULL),
				m_takes_value(true),
				m_needs_value(true),
				m_required(false),
				m_set(false),
				m_option_type(type) {};
	*/
		ZiprOption_t(ZiprOptionType type, std::string key, std::string value)
			: m_key(key),
			  m_untyped_value(value),
				m_observee(NULL),
				m_takes_value(true),
				m_needs_value(true),
				m_required(false),
				m_set(false),
				m_option_type(type) {};

		virtual void SetValue(const std::string &value) = 0;
		virtual void Set() = 0;

		std::string StringValue() {
			return m_untyped_value;
		}
		std::string Namespace() {
			return m_namespace;
		}
		std::string Key() {
			return m_key;
		}
		void SetNeedsValue(bool needs_value) {
			m_needs_value = needs_value;
		}
		bool NeedsValue() {
			return m_needs_value;
		}
		void SetTakesValue(bool takes_value) {
			m_takes_value = takes_value;
		}
		bool TakesValue() {
			return m_takes_value;
		}
		
		bool Required() {
			return m_required;
		}
		void SetRequired(bool required) {
			m_required = required;
		}
		bool RequirementMet() {
			return (!m_required) || (m_set);
		}

		virtual std::string Description() {
			return m_description;
		}
		

		void SetDescription(std::string description) {
			m_description = description;
		}

		void AddObserver(ZiprOption_t *observer) {
			assert(observer->m_option_type == m_option_type);
			//observer->SetValue(m_untyped_value);
			observer->m_observee = this;
			m_observers.insert(observer);
		}
	protected:
		std::string m_namespace, m_key, m_untyped_value, m_description;
		std::set<ZiprOption_t*> m_observers;
		ZiprOption_t *m_observee;
		bool m_takes_value, m_needs_value;
		bool m_required;
		bool m_set;
		void SetObserverValues(std::string value) {
			std::set<ZiprOption_t*>::const_iterator it = m_observers.begin();
			std::set<ZiprOption_t*>::const_iterator it_end = m_observers.end();
			for (; it!=it_end; it++)
				(*it)->SetValue(value);
		}
		ZiprOptionType m_option_type;
};
template <typename T>
class ZiprTypedOption_t : public ZiprOption_t
{
	public:
		ZiprTypedOption_t(ZiprOptionType type,
			std::string key,
			std::string value)
			: ZiprOption_t(type, key, value) {}
		void SetValue(const std::string &value) {
			m_set = true;
			m_untyped_value = value;
			m_value = ConvertToTyped(value);
		}
		virtual void Set() {
		}

		T Value() const {
			if (m_observee!=NULL)
			{
				ZiprTypedOption_t<T> *typed_observee =
					static_cast<ZiprTypedOption_t<T>*>(m_observee);
				return typed_observee->Value();
			}
			return m_value;
		}
		operator T() const {
			return Value();
		};
		virtual std::string Description() {
			return std::string(
			/* + ((!m_required)?"[ ":"")
			+ */m_key + ": " 
			+ ((!m_needs_value)?"[":"")
			+ "<" + TypeName() + ">"
			+ ((!m_needs_value)?"]":"")
			+ ((m_description.length())?":\t":"") + m_description 
			/*+ ((!m_required)?" ]":"")*/);
		}
	protected:
		T m_value;
		virtual T ConvertToTyped(std::string) = 0;
		virtual std::string ConvertToUntyped(T) = 0;
		virtual std::string TypeName() = 0;
};

template <class T>
class ZiprCompositeOption_t : public ZiprTypedOption_t<T>, public T
{
	public:
		ZiprCompositeOption_t(ZiprOptionType type,
			std::string key,
			std::string value)
			: ZiprTypedOption_t<T>(type, key, value) {}
		using ZiprTypedOption_t<T>::ConvertToTyped;
		void SetValue(const std::string &value) {
			ZiprTypedOption_t<T>::SetValue(value);
			T::operator=(ConvertToTyped(value));
		}
};

class ZiprStringOption_t : public ZiprCompositeOption_t<std::string>
{
	public:
		ZiprStringOption_t(std::string key, std::string value = "")
			: ZiprCompositeOption_t(ZiprStringOptionType, key, value) {
			SetValue(value);
		}
	protected:
		std::string ConvertToTyped(std::string value_to_convert) {
			return std::string(value_to_convert);
		}
		std::string ConvertToUntyped(std::string value_to_convert) {
			return std::string(value_to_convert);
		}
		std::string TypeName() {
			return "string";
		}
};

class ZiprBooleanOption_t : public ZiprTypedOption_t<bool>
{
	public:
		ZiprBooleanOption_t(std::string key, std::string value = "")
			: ZiprTypedOption_t(ZiprBooleanOptionType, key, value) {
			m_untyped_value = value;
			m_needs_value = false;
		}
		ZiprBooleanOption_t(std::string key, bool value)
			: ZiprTypedOption_t(ZiprBooleanOptionType, key, "") {
			m_value = value;
			m_needs_value = false;
			m_untyped_value = ConvertToUntyped(value);
		}
		void Set() {
			m_untyped_value = "true";
			m_value = true;
			m_set = true;
		}

	private:
		bool ConvertToTyped(std::string value_to_convert) {
			if (value_to_convert == "true")
				return true;
			else
				return false;
		}
		std::string ConvertToUntyped(bool value_to_convert) {
			return (value_to_convert ? std::string("true") : std::string("false"));
		}
		std::string TypeName() {
			return "bool";
		}
};

class ZiprIntegerOption_t : public ZiprTypedOption_t<int>
{
	public:
		ZiprIntegerOption_t(std::string key, std::string value = "")
			: ZiprTypedOption_t(ZiprIntegerOptionType, key, value) {
			m_value = ConvertToTyped(value);
		}
		ZiprIntegerOption_t(std::string key, int value)
			: ZiprTypedOption_t(ZiprIntegerOptionType, key, "") {
			m_value = value;
			m_untyped_value = ConvertToUntyped(value);
		}
		void SetValue(int value) {
			m_value = value;
			m_untyped_value = ConvertToUntyped(value);
		}
	private:
		int ConvertToTyped(std::string value_to_convert) {
			int converted = 0;
			char *endptr = NULL;
			converted = strtol(value_to_convert.c_str(),
			                   &endptr, 10);
			if (*endptr != '\0')
			{
				m_set = false;
				m_untyped_value = "";
				return 0;
			}
			return converted;
		}
		std::string ConvertToUntyped(int value_to_convert) {
			std::stringstream ss;
			ss << value_to_convert;
			return ss.str();
		}
		std::string TypeName() {
			return "integer";
		}
};

class ZiprDoubleOption_t : public ZiprTypedOption_t<double>
{
	public:
		ZiprDoubleOption_t(std::string key, std::string value = "")
			: ZiprTypedOption_t(ZiprDoubleOptionType, key, value) {
			m_value = ConvertToTyped(value);
		}
		ZiprDoubleOption_t(std::string key, double value)
			: ZiprTypedOption_t(ZiprDoubleOptionType, key, "") {
			m_value = value;
			m_untyped_value = ConvertToUntyped(value);
		}
		void SetValue(double value) {
			m_value = value;
			m_untyped_value = ConvertToUntyped(value);
		}
	private:
		double ConvertToTyped(std::string value_to_convert) {
			double converted = 0.;
			char *endptr = NULL;
			converted = strtof(value_to_convert.c_str(),
			                   &endptr);
			if (*endptr != '\0')
			{
				m_set = false;
				m_untyped_value = "";
				return 0;
			}
			return converted;
		}
		std::string ConvertToUntyped(double value_to_convert) {
			std::stringstream ss;
			ss << value_to_convert;
			return ss.str();
		}
		std::string TypeName() {
			return "double";
		}
};
class ZiprOptionsNamespace_t : public std::set<ZiprOption_t*>
{
	public:
		ZiprOptionsNamespace_t(std::string ns) : m_namespace(ns) {}
		std::string Namespace() {
			return m_namespace;
		};
		void PrintNamespace();
		ZiprOption_t *OptionByKey(std::string key);
		void AddOption(ZiprOption_t *option);
		void PrintUsage(int tabs, std::ostream &out);
		bool RequirementsMet();
		void MergeNamespace(ZiprOptionsNamespace_t*);
	private:
		std::string m_namespace;
};

class ZiprOptions_t {
	public:
		ZiprOptions_t(int argc, char **argv);
		void AddNamespace(ZiprOptionsNamespace_t *);
		void PrintNamespaces();
		bool Parse(std::ostream *error=&std::cout, std::ostream *warn=&std::cerr);
		ZiprOptionsNamespace_t *Namespace(std::string);
		void PrintUsage(std::ostream &out);
		bool RequirementsMet();
	private:
		std::vector<std::string> m_arguments;
		std::set<ZiprOptionsNamespace_t*> m_namespaces;
};

}
#endif
